# file: Router.rb
# note: source code is indented with tabs, tab-width=2
#
# A simple rubberband topological router based on the region concept mentioned in
# Tal Dayan's PhD thesis and other papers. 
# Currently we use random points as terminals and try to connect pairs of them.
# Currently we only generate a PGN picture (lines and arcs) using cairo.
#
# c Stefan Salewsk, mail@ssalewski.de
# License GPL
# Version 0.02 02-MAR-2013
#
# Call: ruby router.rb
# or ruby router.rb seed
# for reproduceable results
#
require_relative '../RBOOST/rboost_fibonacci_queue'
require_relative '../RCGAL/rcgal_cdt'

require_relative 'canvas'

Board_Size = 800 # cairo PNG picture
Points = 64 # terminal count

# defaults for testing -- of course later each pin/trace can have it's own values
# currently we use pixes as size unit, later maybe nm (int) or m (float)
Pin_Radius = 10
Trace_Width = 7
Clearance = 4
MinCutSize = 50

	def init_seed
		seed = (ARGV[0] ? ARGV[0].to_i : rand(1000))
		print 'random seed is ', seed, "\n"
		srand(seed)
		seed
	end

	def unused_inner_angle(o, a, b)
		ax = a.x - o.x
		ay = a.y - o.y
		bx = b.x - o.x
		by = b.y - o.y
		d = (Math.atan2(by, bx) - Math.atan2(ay, ax)).abs
		d =  2 * Math::PI - d if d > Math::PI
		d
	end

# Net connecting two terminals (2Net)
class NetDesc
	@@id = 0
	attr_accessor :name1, :name2 # name of terminals
	attr_accessor :thickness
	attr_accessor :separation
	attr_accessor :id
	attr_accessor :flag # used for sorting
	def initialize(name1, name2, thickness = Trace_Width, separation = Clearance)
	  @id = @@id += 1
		@name1, @name2 = name1, name2
		@thickness, @separation = thickness, separation
	end
end

class NetDescList < Array
end

# Incident or attached net of a terminal
class Step
	attr_accessor :id # id of this net
	attr_accessor :net_desc
	attr_accessor :vertex # the terminal we are attached or incident to
	attr_accessor :prev, :next # terminal
	attr_accessor :pstep, :nstep # step
	attr_accessor :a, :b, :d, :g, :dir
	attr_accessor :radius
	attr_accessor :score # for sorting attached nets of a terminal by angle 
	attr_accessor :index
	attr_accessor :ref
	attr_accessor :rgt
	attr_accessor :convex


	def initialize(prev, nxt, id)
		@prev, @next, @id = prev, nxt, id
		@radius = 0 # default for incident net
		@convex = false
	end

	# maybe we should cache this -- should not change
	def cross
		(@prev.x - @vertex.x) * (@next.y - @vertex.y) - (@prev.y - @vertex.y) * (@next.x - @vertex.x)
	end

	def inner_angle 
		d = (Math.atan2(@next.y - @vertex.y , @next.x - @vertex.x) - Math.atan2(@prev.y - @vertex.y, @prev.x - @vertex.x)).abs
		d > Math::PI ? 2 * Math::PI - d : d
	end
end

# Terminal
class Vertex < CGAL::Vertex
	@@id = 0
	attr_accessor :id
	attr_accessor :vis_flag # for debugging
	attr_accessor :radius # outer copper
	attr_accessor :separation # outer clearance
	attr_accessor :neighbors # Vertex/Terminal, the neighbors in the delaunay triangulation
	attr_accessor :incident_nets # Step, nets incident to this terminal
	attr_accessor :attached_nets # Step, nets attached to this terminal
	attr_accessor :name # name of the Terminal, i.e. U7_3 -- pin 3 of IC U7
	attr_accessor :tradius, :trgt # temporary data

	def initialize(x = 0, y = 0)
		super(x, y)
		@vis_flag = 0
	  @id = @@id
		@@id += 1
		@radius, @separation = Pin_Radius, Clearance
		@name = ''
		#@x = x.to_f # this to_f makes a difference in net ordering (104) -- we have to investigate why
		#@y = y.to_f # but only when we define "def eql?(other)" below -- I have no idea why
		@neighbors = Array.new
		@incident_nets = Array.new
		@attached_nets = Array.new
	end

  def reset_initial_size
		@radius, @separation = Pin_Radius, Clearance
	end

	# may be obsolete -- at least it is only an estimation
	def resize
		reset_initial_size
		attached_nets.each{|step|
			net = step.net_desc
			trace_sep = [@separation, net.separation].max
			@radius += trace_sep + net.thickness
			step.radius = @radius - net.thickness * 0.5
			@separation = net.separation
		}
	end

	# may be obsolete -- at least it is only an estimation
  def update(s)
		net = s.net_desc
		trace_sep = [@separation, net.separation].max
		@radius += trace_sep + net.thickness
		s.radius = @radius - net.thickness * 0.5
		@separation = net.separation
	end

	# unused when equ? is not used
#	def hash
#    [@x, @y].hash
#  end

	# this will break net ordering -- at least with @x = x.to_f above
#  def eql?(other)
#    [@x, @y].eql?([other.x, other.y])
#  end

	# returns step -- may become obsolete
  def net(id)
		incident_nets.each{|s| return s if s.id == id}
 		attached_nets.each{|s| return s if s.id == id}
		return nil
	end

  # delete step -- currently we need this, but we should try to avoid it, at least the expensive resize operation
	def delete_net(id)
		incident_nets.delete_if{|step| step.id == id}
		attached_nets.delete_if{|step| step.id == id}
		resize
	end

	# 2 PI < 7 < 8
	def _full_angle(s, group_sign, new_cross)
		#return (8 + s.net_desc.id) * new_cross unless s.next && s.prev # incident nets outer most
		return Math::PI unless s.next && s.prev # assume straight continuation -- problem is more than one, maybe + s.net_desc.id * epsilon
		d = (Math.atan2(s.prev.y - s.vertex.y, s.prev.x - s.vertex.x) - Math.atan2(s.next.y - s.vertex.y, s.next.x - s.vertex.x)) * s.net_desc.flag * group_sign
		d < 0 ? 2 * Math::PI + d : d
	end

	# this is the initially version with two hashes, which allows better detection of contradictions. Do not remove, we may need as fallback.
	def unused_sort_attached_nets # by angle of total trace
		unless attached_nets.length < 2
			attached_nets.each{|n| fail unless n.vertex == self; n.index = -n.inner_angle} # largest angle inner most
			attached_nets.sort_by!{|n| n.index}
			attached_nets.each_with_index{|n, i| n.index = i}
			shash = Hash.new
			attached_nets.each{|n| # group attached nets with same angle (overlapping)
				l = n.prev
				r = n.next
				n.net_desc.flag = 1
				if shash.has_key?([l, r])
					shash[[l, r]] << n
				elsif shash.has_key?([r, l])
					n.net_desc.flag = -1 # inverted direction
					shash[[r, l]] << n
				else
					shash[[l, r]] = [n]
				end
			}
			shash.each_value{|group| # fine sort each group by following the traces
				if group.length > 1
					group.each{|el| el.ref = el}
					indices = Array.new
					group.each{|el| indices << el.index}
					indices.sort!
					rel = Hash.new#(0) # we use two hashes to detect contradictions -- if all is fine we can use only one common
					rul = Hash.new#(0) # (0) is only needed if we can not really decide order
					group_sign = (group[0].cross <=> 0)
					new_cross = 1
					gr = group.dup
					while gr.length > 1
						gr.map!{|el| (el.net_desc.flag == group_sign ? el.pstep : el.nstep)} # walk in one direction
						gr.each{|el| el.ref = (el.net_desc.flag == group_sign ? el.nstep.ref : el.pstep.ref)}
						### new_cross = nil
						gr.each{|el| if el.next && el.prev then new_cross = (0 <=> el.cross * el.ref.cross); break; end}
						### fail unless new_cross # all remaining nets stop here -- we assign a preliminary score, which
						### we may overwrite when we walk in the other direction
						gr.each{|el| el.score = _full_angle(el, group_sign, new_cross)}
						unresolved_combinations = false
						gr.combination(2).each{|el|
							a, b = *el
							unless rel.has_key?([a.ref, b.ref])
								if (c = (a.score <=> b.score)) != 0
								  if  a.score < 7 ||  b.score < 7 # indicate final sort
										c *= 2
									end
									rel[[a.ref, b.ref]] = c
									rel[[b.ref, a.ref]] = -c
								else
								  unresolved_combinations = true
								end
							end
						}
						break unless unresolved_combinations
						gr.keep_if{|el| el.next && el.prev}
					end
					gr = group.dup
					while gr.length > 1
						gr.map!{|el| (el.net_desc.flag == group_sign ? el.nstep : el.pstep)} # walk in one direction -- the other
						gr.each{|el| el.ref = (el.net_desc.flag == group_sign ? el.pstep.ref : el.nstep.ref)}
						### new_cross = nil
						gr.each{|el| if el.next && el.prev then new_cross = (0 <=> el.cross * el.ref.cross); break; end}
						### fail unless new_cross # all remaining nets stop here -- we assign a preliminary score, which
						### we may overwrite when we walk in the other direction
						gr.each{|el| el.score = _full_angle(el, group_sign, new_cross)}
						unresolved_combinations = false
						gr.combination(2).each{|el|
							a, b = *el
							unless rul.has_key?([a.ref, b.ref])
								if (c = (a.score <=> b.score)) != 0
								  if  a.score < 7 ||  b.score < 7 # indicate final sort
										c *= 2
									end
									rul[[a.ref, b.ref]] = c
									rul[[b.ref, a.ref]] = -c
								else
								  unresolved_combinations = true
								end
							end
						}
						break unless unresolved_combinations
						gr.keep_if{|el| el.next && el.prev}
					end
					unless rel.eql?(rul)
						puts 'rel and rul differ'
						rel.each_pair{|k, v|
							if rul.has_key?(k) and (rul[k] != v)
								fail if v.abs == 2 && rul[k].abs == 2 # contradicting final relation
								puts k, v, rul[k] # they may differ when the net terminates early for at least one direction -- score.abs > 2 PI
							end
						}
					end
					# nets b and c end at the same terminal for left side -- so we have to investigate the other side for the
					# correct final ordering. 
					#   \              /
					# a  --------------
					# b    --------------------
					# c    --------------
					#                    \
					# for fully parallel nets arbitrary well defined ordering should work -- score is (8 + s.net_desc.id) for that case
					# -------------------
					#   ------------
					# it seems that the block in merge! is executed always, even when v1 == v2 -- bug in Ruby 1.9.3?
					rel.merge!(rul){|k, v1, v2| v1.abs > v2.abs ? v1 : v2} # select the final value
					group.sort!{|a, b| rel[[a, b]]} # do we need rel[[a, b] <=> 0 to ensure -1,0,1 in block? 
					group.each{|el| el.index = indices.shift}
				end
			}
			attached_nets.sort_by!{|el| el.index}
		end
	end

	def sort_attached_nets # by angle of total trace
		unless attached_nets.length < 2
			attached_nets.each{|n| fail unless n.vertex == self; n.index = -n.inner_angle} # largest angle inner most
			attached_nets.sort_by!{|n| n.index}
			attached_nets.each_with_index{|n, i| n.index = i}
			shash = Hash.new
			attached_nets.each{|n| # group attached nets with same angle (overlapping)
				l = n.prev
				r = n.next
				n.net_desc.flag = 1
				if shash.has_key?([l, r])
					shash[[l, r]] << n
				elsif shash.has_key?([r, l])
					n.net_desc.flag = -1 # inverted direction
					shash[[r, l]] << n
				else
					shash[[l, r]] = [n]
				end
			}
			shash.each_value{|group| # fine sort each group by following the traces
				if group.length > 1
					group.each{|el| el.ref = el}
					indices = Array.new
					group.each{|el| indices << el.index}
					indices.sort!
					rel = Hash.new
					group_sign = (group[0].cross <=> 0)
					new_cross = 1
					[-1, 1].each{|direction|
						gr = group.dup
						final = true
						while gr.length > 1
							gr.map!{|el| (el.net_desc.flag == direction ? el.pstep : el.nstep)} # walk in one direction
							gr.each{|el| el.ref = (el.net_desc.flag == direction ? el.nstep.ref : el.pstep.ref)}
							### new_cross = nil
							gr.each{|el| if el.next && el.prev then new_cross = (0 <=> el.cross * el.ref.cross); break; end}
							### fail unless new_cross # all remaining nets stop here -- we assign a preliminary score, which
							### we may overwrite when we walk in the other direction
							gr.each{|el| el.score = _full_angle(el, group_sign, new_cross)}
							unresolved_combinations = false
							gr.combination(2).each{|el|
								a, b = *el
								relation = rel[[a.ref, b.ref]]
								if !relation || relation.abs < 2
									if (c = (a.score <=> b.score)) != 0
										if  a.score < 7 ||  b.score < 7 # indicate final relation
											c *= 2
										else
											final = false
										end
										#if !relation || (relation.abs < c.abs)
											rel[[a.ref, b.ref]] = c
											rel[[b.ref, a.ref]] = -c
										#end
									else
										unresolved_combinations = true
									end
								end
							}
							break unless unresolved_combinations
							gr.keep_if{|el| el.next && el.prev}
						end
						fail if unresolved_combinations # we should get at least a preliminary relation
						break if final
					}
					group.sort!{|a, b| rel[[a, b]]} # do we need rel[[a, b] <=> 0 to ensure -1,0,1 in block? 
					group.each{|el| el.index = indices.shift}
				end
			}
			attached_nets.sort_by!{|el| el.index}
		end
	end
end

class Region
	attr_accessor :vertex, :neighbors, :name, :incident
  def initialize(v = nil)
		@vertex = v
		@neighbors = Array.new
		@name = nil
		@incident = true
	end
end

class Cut
  attr_accessor :cap, :flow
  def initialize(dx, dy)
		@cap = Math.sqrt(dx ** 2 + dy ** 2)
		@flow = 0
	end
end

# key is [ax, ay, bx, by] -- coordinates of two points
# should be insensitive when points are exchanged.
=begin
class Cut_Hash < Hash
  def [](k0, k1, k2, k3)
		if k0 < k2
			super([k0, k1, k2, k3])
		elsif k0 == k2
			k1 > k3 ? super([k2, k3, k0, k1]): super([k0, k1, k2, k3])
	  else
			super([k2, k3, k0, k1])
    end
  end

  def []=(k0, k1, k2, k3, v)
		if k0 < k2
			super([k0, k1, k2, k3], v)
		elsif k0 == k2
			k1 > k3 ? super([k2, k3, k0, k1], v): super([k0, k1, k2, k3], v)
	  else
			super([k2, k3, k0, k1], v)
    end
  end
end
=end

class Cut_Hash < Hash
  def [](k0, k1, k2, k3)
		super([[k0, k1], [k2, k3]].sort)
	end
  def []=(k0, k1, k2, k3, v)
		super([[k0, k1], [k2, k3]].sort, v)
	end
end

# we put only one pin in each cell
CS = 30 # Cell_Size
#
class Router
	attr_accessor :filename
  def initialize
	  @Path_ID = 0
		@image = Cairo::ImageSurface.new(Board_Size, Board_Size)
		@pic = Canvas::Pic.new(@image)
		@pic.set_source_rgba(0.8, 0.8, 0.8, 1)
		@pic.paint
		cdt = CGAL::CDT.new
		cell = Hash.new
		# put a virtual pin on each corner of our board -- we should need this
		[0, Board_Size].repeated_permutation(2).each{|el|
			v = Vertex.new(*el)
			cdt.insert(v)
			cell[[el[0] / CS, el[1] / CS]] = 1
		}

# we can use constraints...
=begin
		con = Array.new
		[300, 500].repeated_permutation(2).each{|el|
			v = Vertex.new(*el)
			con << v
			cdt.insert(v)
			el.map!{|x| x / CS}
			cell[el] = 1
		}
		cdt.insert_constraint(con[0], con[1])
		cdt.insert_constraint(con[0], con[2])
		cdt.insert_constraint(con[3], con[1])
		cdt.insert_constraint(con[3], con[2])
=end

		id = 8
		while id < Points
			r1, r2 = rand(Board_Size - 20) + 10, rand(Board_Size - 20) + 10
			if cell[[r1 / CS, r2 / CS]] == nil
				cell[[r1 / CS, r2 / CS]] = 1
				#if true#unless (300..500).include?(r1) or (300..500).include?(r2)
				v = Vertex.new(r1, r2)
				cdt.insert(v)
				id += 1
			end
		end

		@vertices = Array.new # maybe later we do not need this array, we have cdt.each{}
		@regions = Array.new
		cdt.each{|v|
			@vertices << v
			@regions << Region.new(v)
		}
		set = @vertices.select{|el| el.id > 3}.each_slice(2).to_a #.shuffle
		set.sort_by!{|el| (el[0].x - el[1].x) ** 2 + (el[0].y - el[1].y) ** 2}
		@netlist = NetDescList.new
		(0..9).each{|i|
			v1, v2 = *set[i * 3]
			v1.name = i.to_s + 's'
			v2.name = i.to_s + 'e'
			net_desc = NetDesc.new(v1.name, v2.name)
			@netlist << net_desc
		}
		@cuts = Cut_Hash.new
		cdt.each{|v|
			cdt.neighbor_vertices(v).each{|n|
				v.neighbors << n
				@regions[v.id].neighbors << @regions[n.id]
				ax = v.x
				ay = v.y
				bx = n.x
				by = n.y
				@cuts[ax, ay, bx, by] = Cut.new(ax - bx, ay - by)
			}
		}
	end

=begin
	def flag_vertices
		@pic.set_source_rgba(1, 0, 0, 1)
		@pic.set_line_width(1)
		@vertices.each{|v|
			if v.vis_flag != 0
				if v.vis_flag == 1
					@pic.set_source_rgba(1, 0, 0, 1)
				else
					@pic.set_source_rgba(0, 1, 0, 1)
				end
				@pic.new_sub_path
				@pic.set_line_width(1)
				@pic.arc(v.x, v.y, Pin_Radius, 0, 2*Math::PI)
				@pic.fill
				@pic.stroke
			end
		}
	end
=end

	def draw_vertices
		@pic.set_source_rgba(0, 0, 0, 0.3)
		@pic.set_line_width(1)
		@vertices.each{|v|
			@pic.new_sub_path
			@pic.arc(v.x, v.y, Pin_Radius, 0, 2 * Math::PI)
			@pic.fill
			@pic.stroke
			v.neighbors.each{|n|
				@pic.move_to(v.x, v.y)
				@pic.line_to(n.x, n.y)
				@pic.stroke
			}
			#@pic.stroke
		}
		@pic.stroke
		@pic.set_source_rgba(1, 0, 0, 0.7)
		@pic.set_line_width(2)
		@cuts.each_pair{|k, v|
			if v.cap < MinCutSize
				@pic.move_to(*k[0])
				@pic.line_to(*k[1])
			end
		}
		@pic.stroke
	end

	def gen_line(x0, y0, x1, y1, w)
		@pic.set_line_width(w)
		@pic.move_to(x0, y0)
		@pic.line_to(x1, y1)
		@pic.stroke
	end

  def gen_arc(x, y, r, from, to, width)
		@pic.new_sub_path
		@pic.set_line_width(width)
		@pic.arc(x, y, r, from, to)
		@pic.stroke
	end

# http://www.faqs.org/faqs/graphics/algorithms-faq/
# http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html
# int pnpoly(int nvert, float *vertx, float *verty, float testx, float testy)
# {
#   int i, j, c = 0;
#   for (i = 0, j = nvert-1; i < nvert; j = i++) {
#     if ( ((verty[i]>testy) != (verty[j]>testy)) &&
# 	 (testx < (vertx[j]-vertx[i]) * (testy-verty[i]) / (verty[j]-verty[i]) + vertx[i]) )
#        c = !c;
#   }
#   return c;
# }
#
# input: array of vertices
# result: the vertices inside the polygon (or on the border?)
	def vertices_in_polygon(p_vertices, test_vertices)
		res = Array.new
		nm1 = p_vertices.length - 1
		test_vertices.each_with_index{|tp, k|
			ty = tp.y
			i = 0
			j = nm1
			c = false
			while i <= nm1
        if ((((p_vertices[i].y <= ty) && (ty < p_vertices[j].y)) ||
             ((p_vertices[j].y <= ty) && (ty < p_vertices[i].y))) &&
            (tp.x < (p_vertices[j].x - p_vertices[i].x) * (ty - p_vertices[i].y) / (p_vertices[j].y - p_vertices[i].y) + p_vertices[i].x))
					 c = !c
				end
				j = i
				i += 1
			end
			res << tp if c
		}
		res
	end

	#      (c)
	#     /
	#    /     (p)
	#   /
	# (b)
	# see http://www.geometrictools.com/
	#
	def unused_distance_line_segment_point_squared(bx, by, cx, cy, px, py)
		mx = cx - bx
		my = cy - by
		hx = px - bx
		hy = py - by
		t0 = (mx * hx + my * hy).fdiv(mx ** 2 + my ** 2)
		if t0 <= 0
		elsif t0 < 1
			hx -= t0 * mx
			hy -= t0 * my
		else
			hx -= mx
			hy -= my
		end
		return hx ** 2 + hy ** 2
	end

	def unused_vertices_touch_line(x1, y1, x2, y2, line_thickness, line_separation, vertices)
		v_in = Array.new
		vertices.each{|v|
			s = [v.separation, line_separation].max
			if unused_distance_line_segment_point_squared(x1, y1, x2, y2, v.x, v.y) < (v.radius + s + line_thickness * 0.5) ** 2
				v_in << v
			end
		}
		v_in
	end

	# P_IN = 1; P_ON = 0; P_OUT = -1 
	# see http://en.wikipedia.org/wiki/Barycentric_coordinates_%28mathematics%29
	def unused_point_in_triangle(x1, y1, x2, y2, x3, y3, x, y)
		d  =  (y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3)
		l1 = ((y2 - y3) * (x - x3) + (x3 - x2) * (y - y3)) / d
		l2 = ((y3 - y1) * (x - x3) + (x1 - x3) * (y - y3)) / d
		l3 = 1 - l1 - l2
		min, max = [l1, l2, l3].minmax
		if 0 <= min && max <= 1
			0 < min && max < 1 ? P_IN : P_ON
		else
			P_OUT
		end
	end

	def unused_vertices_in_triangle(x1, y1, x2, y2, x3, y3, vertices)
		d  =  (y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3)
		v_in = Array.new
		if d == 0
			puts 'vertices_in_triangle, degenerated'
			return v_in
		end
		y2my3 = (y2 - y3) / d
		x3mx2 = (x3 - x2) / d
		x1mx3 = (x1 - x3) / d
		y3my1 = (y3 - y1) / d
		vertices.each{|v|
			vxmx3 = v.x - x3
			vymy3 = v.y - y3
			l1 = y2my3 * vxmx3 + x3mx2 * vymy3
			l2 = y3my1 * vxmx3 + x1mx3 * vymy3
			min, max = [l1, l2, 1 - l1 - l2].minmax
			if 0 <= min && max <= 1
				v_in << v
			end
		}
		v_in
	end

	# see http://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain
	def unused_hcross(o, a, b)
		(a[0] - o[0]) * (b[1] - o[1]) - (a[1] - o[1]) * (b[0] - o[0])
	end
	def unused_convex_hull(points)
		points.sort!.uniq!
		return points if points.length < 3
		lower = Array.new
		points.each{|p|
			while lower.length > 1 and unused_hcross(lower[-2], lower[-1], p) <= 0 do lower.pop end
			lower.push(p)
		}
		upper = Array.new
		points.reverse_each{|p|
			while upper.length > 1 and unused_hcross(upper[-2], upper[-1], p) <= 0 do upper.pop end
			upper.push(p)
		}
		return lower[0...-1] + upper[0...-1]
	end

=begin
Problem: Fat inner therminal.
Vector U tangent from small therminal to inner large one.
Vector V tangent from inner large one to small therminal.
Cross product will change sign when inner large therminal
becomes larger than about the distance of the both small
outer therminals. One solution: calculate the intersection
point of the two vectors: Equation
U0 + aU = V0 + bV
For the usual case a should be positive. But a becomes negative
when the cross product changes sign due to the blown up inner
terminal size. 
      
    -//---\\
    //     \\
 U / |     | \V
  /   \ __ /  \
 /             \
/O            O_\|

U0 + aU = V0 + bV
-
U0x + aUx = V0x + bVx
U0y + aUy = V0y + bVy
-
Solve for a -- its sign is important
(U0x + aUx - V0x) / Vx = b
U0y + aUy = V0y + Vy * (U0x + aUx - V0x) / Vx
a * (Uy - Ux * Vy / Vx) = V0y + Vy * (U0x - V0x) / Vx - U0y
a * (Uy * Vx - Ux * Vy) = V0y * Vx + Vy * (U0x - V0x) - U0y * Vx
a = ((V0y - U0y) * Vx + Vy * (U0x - V0x)) / (Uy * Vx - Ux * Vy)
=end
	### do not delete this, it shows the history
	def unused_unoptimized_convex_vertices(vertices, prev, nxt)
		fail if prev == nxt
		fail if vertices.uniq!
		fail unless vertices.include?(prev) && vertices.include?(nxt)
		return vertices if vertices.length < 3
		x = prev.x
		y = prev.y
		xx = nxt.x - x
		yy = nxt.y - y
		vertices.sort_by!{|el| (el.x - x) * xx + (el.y - y) * yy}#.uniq!
		h = Array.new
		turn = ->(v; x1, y1, x2, y2, x3, y3, x4, y4, h1, h2, a1) {
			h2, h1 = h[-2, 2]
			x1, y1, x2, y2 = get_tangents(h2.x, h2.y, h2.tradius, h2.trgt, h1.x, h1.y, h1.tradius, h1.trgt)
			x3, y3, x4, y4 = get_tangents(h1.x, h1.y, h1.tradius, h1.trgt, v.x, v.y, v.tradius, v.trgt)
			a1 = (y3 - y1) * (x4 - x3) + (x1 - x3) * (y4 - y3) # a = a1 / a2
			#a2 = (y2 - y1) * (x4 - x3) - (x2 - x1) * (y4 - y3) # this is the cross product of U and V
			#return (a2 < 0) ^ h[1].trgt ^ (a1 / a2 < 0)
			return (a1 < 0) ^ h[1].trgt # funny, we do not need a2 -- no problem with division by zero
		}
		vertices.each{|v|
			while h.length > 1 and turn[v] do h.pop end
			h.push(v)
		}
		#h.delete(prev) # caller may do this
		#h.delete(nxt)
		h
	end

	def convex_vertices(vertices, prev, nxt)
		fail if prev == nxt
		fail if vertices.uniq!
		fail unless vertices.include?(prev) && vertices.include?(nxt)
		return vertices if vertices.length < 3 # or []
		x1 = prev.x
		y1 = prev.y
		x3 = nxt.x - x1
		y3 = nxt.y - y1
		vertices.sort_by!{|el| (el.x - x1) * x3 + (el.y - y1) * y3}#.uniq!
		h = Array.new
		t = Array.new
		vertices.each{|v|
			while !h.empty? do
				h1 = h[-1]
				tangent = get_tangents(h1.x, h1.y, h1.tradius, h1.trgt, v.x, v.y, v.tradius, v.trgt)
				if h.length == 1
					t << tangent
					break
				end
				x1, y1 = t[-1]
				x3, y3, x4, y4  = tangent
				if ((y3 - y1) * (x4 - x3) + (x1 - x3) * (y4 - y3) < 0) ^ h[1].trgt
					h.pop
					t.pop
				else
					t << 	tangent
					break
				end
			end
			h.push(v)
		}
		return h
	end

	# intermediate solution, now obsolete
	def unused_convex_vertices(vertices, prev, nxt, h1, h2)
		vertices.delete(prev)
		vertices.delete(nxt)
		return vertices if vertices.empty?
		h = Hash.new
		if vertices.length > 2
			vertices.permutation(2).each{|el|
				a, b = el
				fail if a == b
				bx, by, ax, ay = *get_tangents(b.x, b.y, b.tradius, true, a.x, a.y, a.tradius, true)
				gen_line(ax, ay, bx, by, 1)
				h[[ax, ay]] = a
				h[[bx, by]] = b
			}
		end
		vertices.each{|el|
			bx, by, ax, ay = *get_tangents(prev.x, prev.y, prev.tradius, prev.trgt, el.x, el.y, el.tradius, el.trgt)
			h[[ax, ay]] = el
			h[[bx, by]] = prev
			gen_line(ax, ay, bx, by, 1)
			bx, by, ax, ay = *get_tangents(el.x, el.y, el.tradius, el.trgt, nxt.x, nxt.y, nxt.tradius, nxt.trgt)
			h[[ax, ay]] = nxt
			h[[bx, by]] = el
			gen_line(ax, ay, bx, by, 1)
		}
		return (unused_convex_hull(h.keys + [h1, h2])-[h1, h2]).map{|el| h[el]}.uniq
	end

	def incident_net_to_neighbors?(vertex, neighbors)
		!(vertex.incident_nets.map{|el| el.next || el.prev} & neighbors).empty?
	end

	def unused_basic_dijkstra(start_node)
		# Nodes that may have neighbours wich can be relaxed further
		active = PriorityQueue.new         
		# Best distances found so far
		distances = Hash.new { 1.0 / 0.0 } 
		# Parent pointers describing shortest paths for all nodes
		parents = Hash.new                 
		# Initialize with start node
		active[start_node] = 0
		until active.empty?
			u, distance = active.delete_min
			distances[u] = distance
			d = distance + 1
			u.neighbours.each do | v |
				next unless d < distances[v] # we can't relax this one
				active[v] = distances[v] = d
				parents[v] = u
			end    
		end
		parents
	end
	
	def dijkstra(start_node, end_node_name)
		q = BOOST::Fibonacci_Queue.new
		distances = Hash.new(Float::INFINITY)
		parents = Hash.new
		fail unless start_node
		q[[start_node, false]] = 0
		cross = false
		while true do
			min, distance = q.pop
			u, last_cross = *min
			if !min or ((u.vertex.name == end_node_name) and (u.incident) )#and parents[parents[[u, last_cross]]])
				break
			end
			distances[min] = distance
			x = u.vertex.x
			y = u.vertex.y
			u.neighbors.each do | v |
			fail unless v
				d = distance + Math.sqrt((v.vertex.x - x) ** 2 + (v.vertex.y - y) ** 2)
				old = parents[[u, last_cross]]
				old = old[0] if old
				if old
					cross = no_cross(old, u, v) > 0
					next unless d < distances[[v, cross]]
					lcuts = bor_list(old, v, u)
					lcuts << old.vertex if ((last_cross != nil) and (cross != last_cross))
					lcap = 1e3
					lcuts.each{|c|
						lcap = [@cuts[u.vertex.x, u.vertex.y, c.x, c.y].cap, lcap].min
					}
					#hhh = lcuts
					#lcap = 0 if incident_net_to_neighbors?(u.vertex, lcuts)
					lcap = 0 if old == v

					next if lcap < MinCutSize
				end
				q[[v, cross]] = distances[[v, cross]] = d
				parents[[v, cross]] = [u, last_cross]
			end    
		end
		return nil unless min
		purents = Hash.new
		purents[end_node_name] = u 
		p = min
		while p
			n =  parents[p]
		  purents[p[0]] = n && n[0]
			p = n
		end
		return purents
	end

	def old_dijkstra(start_node, end_node_name)
		#q = PriorityQueue.new
		q = BOOST::Fibonacci_Queue.new
		distances = Hash.new(Float::INFINITY)
		parents = Hash.new
		fail unless start_node
		q[[start_node, false]] = 0
		#cross = last_cross = false
		cross = false
		while true do
			min, distance = q.pop
			u, last_cross = *min
			#puts 'ddd', min, u
			if !min or ((u.vertex.name == end_node_name) and (u.incident) and parents[parents[[u, last_cross]]])
				#parents[end_node_name] = min
				break
			end
			distances[min] = distance
			x = u.vertex.x
			y = u.vertex.y
			u.neighbors.each do | v |
			fail unless v
				d = distance + Math.sqrt((v.vertex.x - x) ** 2 + (v.vertex.y - y) ** 2)
				old = parents[[u, last_cross]]
				old = old[0] if old
				if old
					cross = no_cross(old, u, v) > 0
					next unless d < distances[[v, cross]]
					lcuts = bor_list(old, v, u)
					lcuts << old.vertex if ((last_cross != nil) and (cross != last_cross))
					lcap = 1e3
					lcuts.each{|c|
						lcap = [@cuts[u.vertex.x, u.vertex.y, c.x, c.y].cap, lcap].min
					}
					hhh = lcuts# + [old, v]
					lcap = 0 if incident_net_to_neighbors?(u.vertex, lcuts)
					lcap = 0 if old == v
					jki = u.vertex.incident_nets.select{|el| el.next == old.vertex ||  el.prev == old.vertex}
					#lcap = 0 unless jki.empty?
					jki.each{|el|
						lcap = 0 if cross != last_cross
						if el.next
							n = el.next
							nn = el.nstep.next
						else
							n = el.prev
							nn = el.pstep.prev
						end
						lcap = 0 if nn && (cross(u.vertex, n, nn) > 0) != cross
					}
					jki = u.vertex.incident_nets.select{|el| el.next == v.vertex ||  el.prev == v.vertex}
					#lcap = 0 unless jki.empty?
					jki.each{|el|
					lcap = 0 if cross != last_cross
						if el.next
							n = el.next
							nn = el.nstep.next
						else
						#lcap = 0
							n = el.prev
							nn = el.pstep.prev
						end
						lcap = 0 if nn && (cross(u.vertex, n, nn) > 0) == cross # condition should be ok
					}
					jki = old.vertex.incident_nets.select{|el| el.next == u.vertex ||  el.prev == u.vertex}
					lcap = 0 unless jki.empty? or (cross == last_cross)
					next if lcap < MinCutSize
				end
				q[[v, cross]] = distances[[v, cross]] = d
				parents[[v, cross]] = [u, last_cross]
			end    
		end
		return nil unless min
		purents = Hash.new
		purents[end_node_name] = u 
		p = min
		while p
			n =  parents[p]
		  purents[p[0]] = n && n[0]
			p = n
		end
		return purents
	end

	def no_cross(o, a, b)
		(a.vertex.x - o.vertex.x) * (b.vertex.y - o.vertex.y) - (a.vertex.y - o.vertex.y) * (b.vertex.x - o.vertex.x)
	end

	#          /neighbor
	#     a   /
	# <-------\
	#       /  \b
	#      /    \
	#return these
	def no_split_neighbor_list(a, b, n)

		l = Array.new
		if !a
			n.neighbors.each{|el|
				if no_cross(n, el, b) < 0
#@pic.save
#				@pic.set_source_rgba(0, 1, 0, 1)
#				gen_line(n.vertex.x, n.vertex.y, el.vertex.x, el.vertex.y, 2)
#@pic.restore
l << el
				end
				}
			l.delete(b)
			return l
		elsif !b
			n.neighbors.each{|el|
				if no_cross(n, a, el) < 0
					l << el
				end
				}
			l.delete(a)
			return l
		end


		aax = n.vertex.x - a.vertex.x
		aay = n.vertex.y - a.vertex.y
		bbx = b.vertex.x - n.vertex.x
		bby = b.vertex.y - n.vertex.y
		as = aay.fdiv(aax)
		bs = bby.fdiv(bbx)
		#l = Array.new
		n.neighbors.each{|el|
		if aax*bby-aay*bbx>0
			if (aax * (el.vertex.y - n.vertex.y) - aay * (el.vertex.x - n.vertex.x) > 0) and (bbx * (el.vertex.y - n.vertex.y) - bby * (el.vertex.x - n.vertex.x) > 0)
				l << el
			end
		else
			if (aax * (el.vertex.y - n.vertex.y) - aay * (el.vertex.x - n.vertex.x) > 0) or (bbx * (el.vertex.y - n.vertex.y) - bby * (el.vertex.x - n.vertex.x) > 0)
				l << el
			end

		end
		}
		l.delete(a)
		l.delete(b)
		return l
	end



	def inner_angle(o, a, b)
		ax = a.x - o.x
		ay = a.y - o.y
		bx = b.x - o.x
		by = b.y - o.y
		d = (Math.atan2(by, bx) - Math.atan2(ay, ax)).abs
		d =  2 * Math::PI - d if d > Math::PI
		d
	end

	def cross(o, a, b)
		(a.x - o.x) * (b.y - o.y) - (a.y - o.y) * (b.x - o.x)
	end

	def cross_tangents(a, b, c, id)
		last_step, cur_step, nxt_step = a.net(id), b.net(id), c.net(id)
		t1 = get_tangents(a.x, a.y, last_step.radius, last_step.rgt, b.x, b.y, cur_step.radius, cur_step.rgt)
		t2 = get_tangents(b.x, b.y, cur_step.radius, cur_step.rgt, c.x, c.y, nxt_step.radius, nxt_step.rgt)
		((t1[2] - t1[0]) * (t2[3] - t2[1]) - (t1[3] - t1[1]) * (t2[2] - t2[0])) > 0
	end


	def atan2_tangents(a, b, id)
		last_step, cur_step = a.net(id), b.net(id)
		t1 = get_tangents(a.x, a.y, last_step.radius, last_step.rgt, b.x, b.y, cur_step.radius, cur_step.rgt)
		#if a.id == 48 || b.id == 48
		#gen_line(t1[0], t1[1], t1[2], t1[3], 9)
		#end
		#t2 = get_tangents(b.x, b.y, cur_step.radius, cur_step.rgt, c.x, c.y, nxt_step.radius, nxt_step.rgt)
		Math.atan2(t1[3] - t1[1], t1[2] - t1[0])
		#Math.atan2(t1[1] - t1[3], t1[0] - t1[2])
		#((t1[2] - t1[0]) * (t2[3] - t2[1]) - (t1[3] - t1[1]) * (t2[2] - t2[0])) > 0
	end







	def bor_list(a, b, n)
		a, b, n = a.vertex, b.vertex, n.vertex
		aax = n.x - a.x
		aay = n.y - a.y
		bbx = b.x - n.x
		bby = b.y - n.y
		l = Array.new
		n.neighbors.each{|el|
		el_x_a = aax * (el.y - n.y) - aay * (el.x - n.x)
		el_x_b = bbx * (el.y - n.y) - bby * (el.x - n.x)
		if aax*bby-aay*bbx>0
			if el_x_a > 0 && el_x_b > 0
				l << el
			end
		else
			if el_x_a < 0 && el_x_b < 0
				l << el
			end
		end
		}
		l.delete(a)
		l.delete(b)
		return l
	end


	def old_bor_list(a, b, n)
		aax = n.vertex.x - a.vertex.x
		aay = n.vertex.y - a.vertex.y
		bbx = b.vertex.x - n.vertex.x
		bby = b.vertex.y - n.vertex.y
		l = Array.new
		n.neighbors.each{|el|
		el_x_a = aax * (el.vertex.y - n.vertex.y) - aay * (el.vertex.x - n.vertex.x)
		el_x_b = bbx * (el.vertex.y - n.vertex.y) - bby * (el.vertex.x - n.vertex.x)
		if aax*bby-aay*bbx>0
			if el_x_a > 0 && el_x_b > 0
				l << el
			end
		else
			if el_x_a < 0 && el_x_b < 0
				l << el
			end
		end
		}
		l.delete(a)
		l.delete(b)
		return l
	end

	def route(net_id)
		net_desc = @netlist[net_id]
		from, to = net_desc.name1, net_desc.name2
		return nil unless from
		start_node = nil
    @regions.each{|r|
		if r.incident
		#print  r.vertex.name, '__', from, "\n"
	    if r.vertex.name == from
        start_node = r
				break
		  end
		end
		}
		#puts net_id
fail unless start_node
		parents = dijkstra(start_node, to)
		unless parents
puts 'dijkstra failed!'
return
		end
		r1 = r2 = nil
		cur = parents[to]
		prv = nil
		@pic.move_to(cur.vertex.x, cur.vertex.y)
		while cur do
			nxt = parents[cur]
			if true#prv && nxt
				ne = no_split_neighbor_list(prv, nxt, cur)
				cap = 1e3
				ne.each{|r|
					if (i = @cuts[cur.vertex.x, cur.vertex.y, r.vertex.x, r.vertex.y].cap) < cap then cap = i end
				}
				ne_comp = ((cur.neighbors - ne) - [prv, nxt])#.uniq
				ne << nxt if nxt
				ne_comp << nxt if nxt
				if r1
					ne.delete(r1)
					ne.delete(r2)
					ne_comp.delete(r1)
					ne_comp.delete(r2)
					ne << r1
					ne_comp << r2
				else
					ne << prv if prv
					ne_comp << prv if prv
				end
				@regions.delete(cur)
				r1 = Region.new(cur.vertex)
				r2 = Region.new(cur.vertex)
				if !cur.incident
				r1.incident = false
				r2.incident = false

				end

				r1.name='r1'
				r2.name='r2'
				@regions << r1 << r2
				cur.neighbors.each{|el|
					el.neighbors.delete(cur)
				}
				ne.each{|el|
					el.neighbors << r1
					r1.neighbors << el
				}
				ne_comp.each{|el|
					el.neighbors << r2
					r2.neighbors << el
				}
			end
			if prv && nxt
			if no_cross(prv, cur, nxt)>0 # > 0
				r1.incident = false
			else
				r2.incident = false
			end
			end
			#if out[cur]
			#	prev = parents[cur]
			#	puts 'out'
			#	if prev
			#		@pic.new_sub_path
			#		@pic.arc(prev.vertex.x, prev.vertex.y, 20, 0, 2 * Math::PI)
			#		@pic.stroke
			#	end
			#end
			prv && @pic.move_to(prv.vertex.x, prv.vertex.y)
			@pic.line_to(cur.vertex.x, cur.vertex.y)
			@pic.stroke
      prv = cur
			cur = nxt
		end
		@pic.stroke
		cur = prv = parents[to]
		p = Array.new
		while cur
			v = cur.vertex
			#v.out = out[prv]
			p << v
			prv = cur
			cur = parents[cur]
		end
		pa = Array.new
		pa << p[-1]
		i = p.length - 2
		while i > 0
			if cross(p[i - 1], p[i], p[i + 1])
				pa << p[i]
			else
				pa << p[i]
			end
			i -= 1
		end
		pa << p[0]
		path = Array.new
		pa.each_index{|i|  
			cur = pa[i]
			nxt = (i == pa.length - 1 ? nil : pa[i + 1])
			prv = (i == 0 ? nil : pa[i - 1])
			if prv and nxt
				path << pa[i]
			else
				path << pa[i]
			end
		}
		pstep = nil
		path.each_index{|i|  
			cur = path[i]
			nxt = (i == path.length - 1 ? nil : path[i + 1])
			prv = (i == 0 ? nil : path[i - 1])
			step = Step.new(prv , nxt, @Path_ID)
			step.net_desc = net_desc
			step.vertex = cur
			step.pstep = pstep
			pstep = step
			if prv and nxt
				cur.update(step)

				#step.lft = (cross(prv, cur, nxt) > 0)
				#step.rgt = cur.out
#step.rgt =	!step.lft
step.rgt = cross(prv, cur, nxt) <= 0

				cur.attached_nets << step
			else
				cur.incident_nets << step
			end
		}
		@Path_ID += 1

		s = pstep
		s.nstep = nil
		while p = s.pstep
			p.nstep = s
			s = p
		end
	end

	# see http://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Tangents_between_two_circles
	def get_tangents(x1, y1, r1, l1, x2, y2, r2, l2)
		d_sq = (x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)
		if (d_sq <= (r1-r2)*(r1-r2))
			puts x1, y1, r1, x2, y2, r2
			fail
		end
		d = Math.sqrt(d_sq)
		vx = (x2 - x1) / d
		vy = (y2 - y1) / d
		sign2 = sign1 = 1
		sign2 = (l2  ? -1 : 1)
		sign1 = (l1  ? -1 : 1)
		sign2 *= sign1
		c = (r1 - sign2 * r2) / d
		#if (c*c > 1.0) continue;
		h = Math.sqrt([0.0, 1.0 - c*c].max)
		nx = vx * c - sign1 * h * vy
		ny = vy * c + sign1 * h * vx
		a = Array.new(4)
		a[0] = x1 + r1 * nx
		a[1] = y1 + r1 * ny
		a[2] = x2 + sign2 * r2 * nx
		a[3] = y2 + sign2 * r2 * ny
		return a
	end

	def sym_get_tangents(x1, y1, r1, l1, x2, y2, r2, l2)
		d_sq = (x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)
		if (d_sq <= (r1-r2)*(r1-r2)) then return nil end
		d = Math.sqrt(d_sq)
		vx = (x2 - x1) / d
		vy = (y2 - y1) / d
		sign2 = sign1 = 1
		sign2 = (l2  ? -1 : 1)
		sign1 = (l1  ? -1 : 1)
		sign2 *= sign1
		c = (r1 - sign2 * r2) / d
		#if (c*c > 1.0) continue;
		h = Math.sqrt([0.0, 1.0 - c*c].max)
		nx = vx * c - sign1 * h * vy
		ny = vy * c + sign1 * h * vx
		a = Array.new(4)
		a[0] = x1 + r1 * nx
		a[1] = y1 + r1 * ny
		a[2] = x2 + sign2 * r2 * nx
		a[3] = y2 + sign2 * r2 * ny
		return a
	end

	def replace(vertex, id, list)
		if list.empty?
			step = vertex.net(id)
 			step.next.net(id).prev = step.prev
 			step.prev.net(id).next = step.next
			step.pstep.nstep = step.nstep
			step.nstep.pstep = step.pstep
		else

		#puts 'replace_insert'
		  pstep = nil
			list.each_index{|i|
				n = Step.new(list[i - 1], list[i + 1], id)
				n.convex = true
				n.net_desc = @netlist[id]
				#puts 'id', id
				fail unless n.net_desc
				n.vertex = list[i]
				n.pstep = pstep
				pstep = n
				n.rgt = !vertex.net(id).rgt
				if i == 0
					n.prev = vertex.net(id).prev
					n.prev.net(id).next = list[0]
					n.pstep = vertex.net(id).pstep
					#n.prev.net(id).nstep = list[0]
				end
				if i == list.length - 1
					n.next = vertex.net(id).next
					n.next.net(id).prev = list[i]
					n.next.net(id).pstep = n
					n.nstep = vertex.net(id).nstep

				end
				#list[i].append_step(n)
				fail unless n.net_desc
				list[i].update(n)
				list[i].attached_nets << n
			}

		i = list.length
		s = pstep
		#s.nstep = nil
		while i > 0
		  i -= 1
		  p = s.pstep
			p.nstep = s
			s = p
		end

		end
		vertex.delete_net(id)
	end

	def init_vertices
		@vertices.each{|vert|
			vert.radius = Pin_Radius 
			vert.separation = Clearance 
		}
	end

	# sort attached nets and calculate its radii
	def no_prepare_steps
		@vertices.each{|vert|
			#vert.attached_nets.sort!{|a, b| inner_angle(a.prev, vert, a.next) <=> inner_angle(b.prev, vert, b.next)}
			#last_sep = vert.separation
			#sum = vert.radius
			
			vert.reset_initial_size
			vert.attached_nets.each{|step|

			net = step.net_desc

			trace_sep = [vert.separation, net.separation].max
			vert.radius += trace_sep + net.thickness
			step.radius = vert.radius - net.thickness * 0.5
			vert.separation = net.separation

			  #net = @netlist[step.id]
				#s = [net.separation, last_sep].max
				#step.radius = sum + s + net.thickness * 0.5
				#sum += s + net.thickness
				#last_sep = net.separation
			}
		}
	end

	#\   |   /
	# \  |  /
	#  \ | /  3 attached concave nets not overlapping
	#    O ------------
	#     \ -----
	#      \
	#       \
	#        \
	# sort attached nets and calculate its radii.
	# this should work for concave (before attachment operator is applied) and convex final nets
	# regard groups by angle: overlapping angles needs different radii for arcs,
	# non overlapping attached arcs may have the same radius.
	# generally one terminal should have at least two final convex groups -- but a terminal with
	# a very big radius may have more than 2 attached groups.
	# Indeed nets never cross, so overlapping is full including
	# currently we collect all attached nets of a terminal in one single array called
	# attached_nets -- we may consider using a separate array for each group...
	# maybe we should regard minimal gap as overlapping? Introduce an eps?
	def smart_prepare_steps
		@vertices.each{|vert|
			next if vert.attached_nets.empty?
			vert.attached_nets.each{|s|
				s.g = -1
				#a = Math.atan2(s.next.y - s.vertex.y, s.next.x - s.vertex.x) # basic angle
				#b = Math.atan2(s.prev.y - s.vertex.y, s.prev.x - s.vertex.x)
				a = atan2_tangents(s.vertex, s.next, s.id) # -PI <= angle <= PI
				b = atan2_tangents(s.prev, s.vertex, s.id) # wrong direction, we have to flip
				b < 0 ? b += Math::PI : b -= Math::PI
				d = (a - b).abs
				dir = (a + b) * 0.5
				if  d > Math::PI # select inner angle < PI -- and flip direction
					d = 2 * Math::PI - d
					dir < 0 ? dir += Math::PI : dir -= Math::PI
				end
				s.d = d * 0.5
				s.dir = dir
			}
			nets = vert.attached_nets.sort_by{|s| s.dir}
			i = nets.length
			last = nets[0]
			last.dir += 2 * Math::PI
			last.g = group = 0
			while (i -= 1) > 0 do # backward while in group 0, start with largest angle
				nxt = nets[i]
				break if nxt.dir + nxt.d < last.dir - last.d
				nxt.g = 0
				last = nxt
			end
			i = 0
			last = nets[0]
			last.dir -= 2 * Math::PI # undo above fix
			while (i += 1) < nets.length do # forward
				nxt = nets[i]
				break if nxt.g == 0
				if nxt.dir - nxt.d > last.dir + last.d
					group += 1
				end
				nxt.g = group
				last = nxt
			end
			group = 0
			loop do
				vert.reset_initial_size
				group_found = false
				vert.attached_nets.each{|step|
					if step.g == group
						group_found = true
						net = step.net_desc
						trace_sep = [vert.separation, net.separation].max
						vert.radius += trace_sep + net.thickness
						step.radius = vert.radius - net.thickness * 0.5
						vert.separation = net.separation
					end
				}
				break unless group_found
				group += 1
			end
		}
	end

	def old_smart_prepare_steps
		@vertices.each{|vert|
			next if vert.attached_nets.empty?
			vert.attached_nets.each{|s|
				s.g = -1
				s.a = Math.atan2(s.next.y - s.vertex.y, s.next.x - s.vertex.x)
				s.b = Math.atan2(s.prev.y - s.vertex.y, s.prev.x - s.vertex.x)
				s.a = atan2_tangents(s.vertex, s.next, s.id)
				s.b = atan2_tangents(s.prev, s.vertex, s.id)
				s.b += Math::PI
				if s.b > Math::PI
					s.b -= 2 * Math::PI
				end
				s.b, s.a = [s.a, s.b].sort
			}
			nets = vert.attached_nets
			group = 0
			lamy = ->(step) {
				nets.each{|s|
					unless s.g > -1
						if step.a - step.b > Math::PI
							touch = (s.a > step.a || s.a < step.b) || (s.b > step.a || s.b < step.b)
						else
							touch = (s.a > step.a && s.a < step.b) || (s.b > step.a && s.b < step.b)
						end
						#print touch, ' ', group, "\n"
						if touch
							s.g = group
							lamy[s]
						end
						s.g=0
					end
				}
			}
			group = 0
			nets.each{|net|
			if net.g == -1
				net.g = group
			#end
				lamy[net]
				end
				group += 1
			}
			group = 0
			loop do
				vert.reset_initial_size
				group_found = false
				vert.attached_nets.each{|step|
					if step.g == group
						group_found = true
						net = step.net_desc
						trace_sep = [vert.separation, net.separation].max
						vert.radius += trace_sep + net.thickness
						step.radius = vert.radius - net.thickness * 0.5
						vert.separation = net.separation
					end
				}
				break unless group_found
				group += 1
			end
		}
	end


	# sort attached nets and calculate its radii
	def prepare_steps
		@vertices.each{|vert|
			next if vert.attached_nets.empty?
			vert.reset_initial_size
			vert.attached_nets.each{|step|
				net = step.net_desc
				trace_sep = [vert.separation, net.separation].max
				vert.radius += trace_sep + net.thickness
				step.radius = vert.radius - net.thickness * 0.5
				vert.separation = net.separation
			}
		}
	end




	# sort attached nets and calculate its radii
	def lazy_prepare_steps
		@vertices.each{|vert|
			next if vert.attached_nets.empty?
			[true, false].each{|b|
				vert.reset_initial_size
				vert.attached_nets.each{|step|
					if step.rgt == b
						net = step.net_desc
						trace_sep = [vert.separation, net.separation].max
						vert.radius += trace_sep + net.thickness
						step.radius = vert.radius - net.thickness * 0.5
						vert.separation = net.separation
					end
				}
			}
		}
	end







	def sort_attached_nets
		@vertices.each{|vert| vert.sort_attached_nets}
	end

  def nubly
	  replaced = true
		while replaced do
	  replaced = false

		@vertices.each{|vert|
			vert.attached_nets.reverse_each{|n|
			  id = n.id
				last = n.prev
				nxt = n.next
				cur = vert
				step = n
				if cross_tangents(last, cur, nxt, id) != step.rgt
				replaced = true
					#rep = vertices_in_triangle(last.x, last.y, cur.x, cur.y, nxt.x, nxt.y, @vertices)
					vec_x = nxt.x - last.x
					vec_y = nxt.y - last.y
					hx, hy = vec_x, vec_y
					vec_x, vec_y = -vec_y, vec_x
					if cross(last, nxt, cur) > 0
						vec_x, vec_y = -vec_x, -vec_y
					end
					h3 = Vertex.new(last.x + (hx + vec_x) * 0.5, last.y + (hy + vec_y) * 0.5)
					rep = vertices_in_polygon([last, cur, nxt, h3], @vertices) - [cur, h3]
					#h1 = Vertex.new(last.x + vec_x, last.y + vec_y)
					#h2 = Vertex.new(nxt.x + vec_x, nxt.y + vec_y)

					h1 = [last.x + vec_x, last.y + vec_y]
					h2 = [nxt.x + vec_x, nxt.y + vec_y]


					rep |= [last, nxt]
					#gen_arc(*h1, 15, 0, 6, 5)
					#gen_arc(*h2, 15, 0, 6, 5)
					#gen_arc(h3.x, h3.y, 15, 0, 6, 15)

					net = cur.net(id).net_desc
					rep.each{|v|
					if v == last || v == nxt
						s = v.net(id)
						v.trgt = s.rgt
						v.tradius = s.radius
					else
					#fail unless step ==  v.net(id)
						#s = v.net(id)
						v.trgt = !step.rgt
						#v.trgt = true# (0 <=> cross(last, nxt, cur))


						v.tradius = v.radius + net.thickness * 0.5 + [net.separation, v.separation].max# + 10
						#print '/// ',  v.tradius, ' ',v.radius, ' ', v.separation, ' ', net.thickness, ' ', net.separation, "\n"
					end
					}
#rep.delete(cur)
set_color(1, 1, 0, 1)
#rep.each{|v| gen_arc(v.x, v.y, 10, 0, 5, 5)}
set_color(1, 1, 1, 1)

#rep.uniq!
#
#=begin
					cur_cross_sign = cross(last, nxt, cur) <=> 0
					dont_touch = [last.radius, nxt.radius].max + 20
					rep.keep_if{|el|
 						(cur_cross_sign == (cross(last, nxt, el) <=> 0)) ||
						(unused_distance_line_segment_point_squared(last.x, last.y, nxt.x, nxt.y, el.x, el.y) < (dont_touch + el.radius) ** 2)
 

						
					}

#=end



					rep.delete(cur)
					fail if last == nxt
fail '**********************************' if unused_unoptimized_convex_vertices(rep, last, nxt) != convex_vertices(rep, last, nxt)
					rep = convex_vertices(rep, last, nxt)#, h1, h2)
#rep.each{|v| gen_arc(v.x, v.y, 10, 0, 5, 5)}

#puts 'ccc', rep.index(last) 

					#rep.rotate(rep.index(last)) if rep.index(last)

					rep.delete(last)
					rep.delete(cur)
					rep.delete(nxt)
					#rep.delete(h1)
					#rep.delete(h2)
					if false #(rep.length > 1) && ((rep[0].x - last.x) ** 2 + (rep[0].y - last.y) ** 2) > ((rep[-1].x - last.x) ** 2 + (rep[-1].y - last.y) ** 2)
						rep.reverse!
					end
					replace(cur, id, rep)
				end
			}
		}
		end
	end

	def draw_routes
		set_color(0, 0, 0, 1)
		@vertices.each{|vert|
			vert.incident_nets.each{|n|
				if false#n.next
					@pic.new_sub_path
					@pic.arc(vert.x, vert.y, 40, 0, Math::PI)
					@pic.stroke
				end
				last = vert
				lastx = lasty = nil
				lr = 0
				to = n.next
				to_net = last.net(n.id)
				while to do
					last_net = to_net
					fail if to_net.nstep != to.net(n.id)

				  to_net = to.net(n.id)
					
					radius = to_net.radius
					#puts 'rrrrrrr', radius
					t = get_tangents(last.x, last.y, lr, last_net.rgt, to.x, to.y, radius, to_net.rgt)
					gen_line(*t, 5)
					if lr > 0
					  start_angle = Math.atan2(lasty - last.y, lastx - last.x)
						end_angle = Math.atan2(t[1] - last.y, t[0] - last.x)
						start_angle, end_angle = end_angle, start_angle unless last_net.rgt
						#@pic.set_source_rgba(1, 0, 0, 1) if last_net.convex	
						gen_arc(last.x, last.y, lr, start_angle, end_angle, 5)
						#@pic.set_source_rgba(0, 0, 0, 1) 
					end
					lr = radius
					last = to
					to = to_net.next
					lastx = t[2]
					lasty = t[3]
				end
			}
		}
	end

	def set_line_width(w)
		@pic.set_line_width(w)
	end
	def set_color(r, g, b, a)
		@pic.set_source_rgba(r, g, b, a)
	end
	def save_picture
		@image.write_to_png(@filename)
	end
end

rs = init_seed
r = Router.new
r.filename = 'pic' + rs.to_s.rjust(3, "0") + '.png'
r.draw_vertices

col = ([1, 0, 0].permutation(3).to_a + [1, 1, 0].permutation(3).to_a).uniq - [[1, 0, 0]]
r.set_color(1, 0, 0, 0.7)
r.set_line_width(4)
(0..9).each{|i|
#(2..2).each{|i|
r.set_color(*col[i % 5], 0.4)
r.route(i)
}
r.sort_attached_nets

r.prepare_steps
r.smart_prepare_steps
r.nubly
#r.sort_attached_nets
puts 'fffffffffffffffffffff'
r.smart_prepare_steps

r.nubly
r.draw_routes
#r.flag_vertices
r.save_picture


